/*	Author: A.Voss@FH-Aachen.de
 * 
 * 	Thema Mehrfachvererbung, inheritance
 */

#include <iostream>
#include <cstdlib>
using namespace std;

#include "../../include/tools.h"

class A 
{
protected:
    int n;
    
public: 
    A(int N) : n(N) { tools_log(); cout << "ctor A(" << n << ")" << endl; }
    ~A()            { tools_log(); cout << "dtor A(" << n << ")" << endl; }
    
    int getn() const { return n; }
};

class B1 : virtual public A				/*	virtual public A ?
 * 
 * 	Siehe class C.
 */
{
public:
    B1(int N) : A(N) { tools_log(); cout << "ctor B1(" << n << ")" << endl; }
    ~B1()            { tools_log(); cout << "dtor B1(" << n << ")" << endl; }
};

class B2 : virtual public A
{
public:
    B2(int N) : A(N) { tools_log(); cout << "ctor B2(" << n << ")" << endl; }
    ~B2()            { tools_log(); cout << "dtor B2(" << n << ")" << endl; }
};

class C : public B1, public B2			/*	Mehrfachvererbung ?
 * 
 * 	Mehrfachvererbung: hier erbt C sowohl von B1 als auch von B2; 
 * 	das bedeutet, alle member von B1 und von B2.
 * 	Die Besonderheit ist hier, dass A als virtual in B1 und B2
 * 	gekennzeichnet ist. Das bedeutet, die Member von A werden alle
 * 	"uebereinandergelegt", sie gibt es also nur einmal! Das hat die
 * 	Form eines Diamanten.
 */

{
public:
    C(int N) : A(N), B1(N+1),B2(N+2)	{ /*	A ?
 * 	
 * 	Hier wird explizit der ctor von A aufgerufen, da er sonst in B1 und 
 * 	B2 zweimal aufgerufen wuerde! Insbesondere wird er nicht mit den 
 * 	Werten in B1 und B2 initialisiert.
 * 	Das ist eine Besonderheit, bei der man explizit einen ctor einer Klasse
 * 	aufruft, von der man nicht direkt erbt.
 */
		tools_log(); cout << "ctor C(" << B1::n << "," << B2::n << ")" << endl; 
	}
    ~C()                       			{ tools_log(); cout << "dtor C(" << B1::n << "," << B2::n << ")" << endl; }
};

int main()
{
    tools_log(); cout << "define objects" << endl;

	// Man beachte die ctor-Aufrufe!
    tools_log(); cout << "c:" << endl;
    C c(0);
    
    tools_log(); 
    cout << "size B1 = " << sizeof(B1) 
         << ", size B2 = " << sizeof(B2) 
         << ", size C = " << sizeof(C) << endl;
    
    tools_log(); cout << "end of main" << endl;
    
    return EXIT_SUCCESS;
}

