/*	Author: A.Voss@FH-Aachen.de
 * 
 * 	Verwendung von virtuellen Funktionen. Motivation.
 */

#include <iostream>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

#pragma GCC diagnostic ignored "-Wdelete-non-virtual-dtor"

// Object sei die Basisklasse aller Objekte
class Object {
public:
	virtual string toString() { return "Object::toString"; }
};

// A erbt von Object
class A : public Object
{
public:
    string toString() { return "A::toString"; }
};

// B erbt von Object
class B : public Object
{
public:
    string toString() { return "B::toString"; }
};

void Out(Object * po) {					/*	Object * ?
 * 
 * 	Hier ist die Anforderung, dass alle Objekte, die von Object
 * 	abgeleitet sind, diese Funktion aufrufen koennen und dann
 * 	die jeweilige (richtige?) Funktion toString aufgerufen werden
 * 	soll ...
 */
    tools_log(); cout << "  *-out: " << po->toString() << endl;
}

void Out(Object & o) {					/*	Object &
 *
 * 	Gleiche Intention wie bei Out oben.
 */
    tools_log(); cout << "  &-out: " << o.toString() << endl;
}

int main()
{
    tools_log(); cout << "Call as member functions" << endl;

    Object o;
    A a;
    B b;
    
    // Was passiert?
    tools_log(); cout << "  o:" << o.toString() << endl;
    tools_log(); cout << "  a:" << a.toString() << endl;
    tools_log(); cout << "  b:" << b.toString() << endl << endl;

    tools_log(); cout << "Call via base-ptr/ref" << endl;
    
    // Was passiert?
    Out(&o);
    Out(&a);
    Out(&b);
    cout << endl;

    // Was passiert?
    Out(o);
    Out(a);
    Out(b);
    
    /* 	Was muss man tun, damit das gewuenschte Verhalten eintritt?
     * 
     * 	toString der "eigentlichen" Klasse des Objektes soll
     * 	aufgerufen werden, nicht die Funktion der Basisklasse.
     * 	Dazu muss man die Funktion toString in der Basisklasse mit
     * 		virtual
     * 	kennzeichnen, dann wird der geuwenschte Effekt erreicht.
     */

    return EXIT_SUCCESS;
}

