/*	Author: A.Voss@FH-Aachen.de
 * 
 * 	Verwendung von virtuellen Funktionen. .
 */

#include <iostream>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

// Object sei die Basisklasse aller Objekte
class Object {
public:
	virtual ~Object() { }
	virtual string toString() { return "Object::toString"; }
};

// Eine Klasse mit einer clone-Methode.
class ICloneable {
public:
    virtual Object* clone() = 0;		/*	=0 ?
     * 
     * 	Das ist eine abstrakte Klasse mit einer "pure virtual function" 
     * 	=> eine Art Interface. Gekennzeichnet durch "=0".
     * 	Diese Methode muss in den Klassen definiert werden, die
     * 	von ICloneable ableiten, damit Instanzen der Klassen
     * 	erzeugt werden koennen. Sonst bleibten sie ebenfalls abstrakte 
     * 	Klassen.
     */
};

class A : public Object, public ICloneable
{
public:
    A()		{ tools_log(); cout << "ctor A" << endl; }
    ~A()    { tools_log(); cout << "dtor A" << endl; }

    string toString() { return "A::toString"; }
    Object* clone() { tools_log(); cout << "clone A" << endl; return new A; }
};

class B : public Object, public ICloneable
{
public:
    B()		{ tools_log(); cout << "ctor B" << endl; }
    ~B()    { tools_log(); cout << "dtor B" << endl; }

    string toString() { return "B::toString"; }
    Object* clone() { tools_log(); cout << "clone B" << endl; return new B; }
};

int main()
{
    tools_log(); cout << "define two original objects" << endl;
    A a;
    B b;
        
    tools_log(); cout << "define the vector" << endl;
    vector<Object*> v;					/*	vector of Object* ?
     * 
     * 	Dies ist ein Container von dynamisch angelegten Objekten vom
     * 	Typ A oder B.
     */

	// Muss clone hierfuer virtual sein?
    v.push_back(a.clone());
    v.push_back(b.clone());
    
    // Was passiert hier? 
    // v.push_back( ((ICloneable*)v[0])->clone() );
	v.push_back( dynamic_cast<ICloneable*>(v[0])->clone() );
    
    for (Object* o : v) {
        tools_log(); cout << "  v:" << o->toString() << endl;
    }
    
    for (Object* o : v)					/*	notwendig?
     * 	
     * 	Ist diese Schleife notwendig, oder loescht clear alles sowieso?
     */
        delete o;
    v.clear();

    tools_log(); cout << "end of main" << endl;

    return EXIT_SUCCESS;
}

